@@ -16,10 +16,18 @@
 -- You should have received a copy of the GNU General Public License
 -- along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-{-# LANGUAGE OverloadedStrings, ConstraintKinds, FlexibleContexts, QuasiQuotes #-}
+{-# LANGUAGE OverloadedStrings, ConstraintKinds, FlexibleContexts,
+             QuasiQuotes, RankNTypes, GeneralizedNewtypeDeriving,
+             FlexibleInstances, MultiParamTypeClasses, UndecidableInstances,
+             TypeFamilies #-}
 
 module Network.Protocol.HTTP.DAV (
-    DAVState
+    DAVT(..)
+  , evalDAVT
+  , setCreds
+  , setDepth
+  , setResponseTimeout
+  , setUserAgent
   , DAVContext(..)
   , getProps
   , getPropsAndContent
@@ -29,18 +37,32 @@ module Network.Protocol.HTTP.DAV (
   , moveContent
   , makeCollection
   , caldavReport
+  , caldavReportM
+  , delContentM
+  , getPropsM
+  , getContentM
+  , mkCol
+  , moveContentM
+  , putPropsM
+  , putContentM
+  , withLockIfPossible
+  , withLockIfPossibleForDelete
   , module Network.Protocol.HTTP.DAV.TH
 ) where
 
 import Network.Protocol.HTTP.DAV.TH
 
-import Control.Applicative (liftA2)
+import Control.Applicative (liftA2, Applicative)
 import Control.Exception.Lifted (catchJust, finally, bracketOnError)
-import Control.Lens ((.~), (^.))
-import Control.Monad (when)
-import Control.Monad.Trans (lift)
-import Control.Monad.Trans.Resource (MonadResourceBase, ResourceT, runResourceT, allocate)
-import Control.Monad.Trans.State.Lazy (evalStateT, StateT, get, modify)
+import Control.Lens ((^.), (.=), (%=))
+import Control.Monad (liftM, liftM2, when, MonadPlus)
+import Control.Monad.Base (MonadBase(..))
+import Control.Monad.Error (ErrorT, MonadError, runErrorT)
+import Control.Monad.Fix (MonadFix)
+import Control.Monad.Trans (lift, MonadTrans)
+import Control.Monad.IO.Class (liftIO, MonadIO)
+import Control.Monad.State (evalStateT, get, MonadState, StateT)
+import Control.Monad.Trans.Control (MonadBaseControl(..))
 
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as BC8
@@ -49,7 +71,8 @@ import qualified Data.Map as Map
 
 import Data.Maybe (catMaybes, fromMaybe)
 
-import Network.HTTP.Conduit (httpLbs, parseUrl, applyBasicAuth, Request(..), RequestBody(..), Response(..), newManager, closeManager, ManagerSettings(..), def, HttpException(..))
+import Network.HTTP.Client (RequestBody(..), httpLbs, parseUrl, applyBasicAuth, Request(..), Response(..), newManager, closeManager, HttpException(..))
+import Network.HTTP.Client.TLS (tlsManagerSettings)
 import Network.HTTP.Types (hContentType, Method, Status, RequestHeaders, unauthorized401, conflict409)
 
 import qualified Text.XML as XML
@@ -58,35 +81,56 @@ import Text.Hamlet.XML (xml)
 
 import Data.CaseInsensitive (mk)
 
-type DAVState m a = StateT (DAVContext m) (ResourceT m) a
+newtype DAVT m a = DAVT { runDAVT :: ErrorT String (StateT DAVContext m) a }
+    deriving (Applicative, Functor, Monad, MonadBase b, MonadError String, MonadFix, MonadIO, MonadPlus, MonadState DAVContext)
 
-initialDS :: String -> B.ByteString -> B.ByteString -> Maybe Depth -> ManagerSettings -> IO (DAVContext a)
-initialDS u username password md s = do
-    mgr <- newManager s
-    req <- parseUrl u
-    return $ DAVContext [] req [] mgr Nothing username password md
+instance MonadBaseControl b m => MonadBaseControl b (DAVT m) where
+   newtype StM (DAVT m) a = StDAVT { unStDAVT :: StM (ErrorT String (StateT DAVContext m)) a }
+   liftBaseWith f = DAVT . liftBaseWith $ \r -> f $ liftM StDAVT . r . runDAVT
+   restoreM       = DAVT . restoreM . unStDAVT
 
-closeDS :: DAVContext a -> IO ()
-closeDS = closeManager . _httpManager
+instance MonadTrans DAVT where
+    lift = DAVT . lift . lift
 
-withDS :: MonadResourceBase m => String -> B.ByteString -> B.ByteString -> Maybe Depth -> DAVState m a -> m a
-withDS url username password md f = runResourceT $ do
-    (_, ds) <- allocate (initialDS url username password md def) closeDS
-    evalStateT f ds
+evalDAVT :: MonadIO m => String -> DAVT m a -> m (Either String a)
+evalDAVT u f = do
+    mgr <- liftIO $ newManager tlsManagerSettings
+    req <- liftIO $ parseUrl u
+    r <- (evalStateT . runErrorT . runDAVT) f $ DAVContext [] req B.empty B.empty [] Nothing mgr Nothing "hDav-using application"
+    liftIO $ closeManager mgr
+    return r
 
-davRequest :: MonadResourceBase m => Method -> RequestHeaders -> RequestBody (ResourceT m) -> DAVState m (Response BL.ByteString)
+choke :: IO (Either String a) -> IO a
+choke f = do
+   x <- f
+   case x of
+       Left e -> error e
+       Right r -> return r
+
+setCreds :: MonadIO m => B.ByteString -> B.ByteString -> DAVT m ()
+setCreds u p = basicusername .= u >> basicpassword .= p
+
+setDepth :: MonadIO m => Maybe Depth -> DAVT m ()
+setDepth d = depth .= d
+
+setUserAgent :: MonadIO m => B.ByteString -> DAVT m ()
+setUserAgent ua = userAgent .= ua
+
+setResponseTimeout :: MonadIO m => Maybe Int -> DAVT m ()
+setResponseTimeout rt = baseRequest %= \x -> x { responseTimeout = rt }
+
+davRequest :: MonadIO m => Method -> RequestHeaders -> RequestBody -> DAVT m (Response BL.ByteString)
 davRequest meth addlhdrs rbody = do
     ctx <- get
     let hdrs = catMaybes
-               [ Just (mk "User-Agent", (BC8.pack ("hDav-using application")))
+               [ Just (mk "User-Agent", ctx ^. userAgent)
                , fmap ((,) (mk "Depth") . BC8.pack . show) (ctx ^. depth)
                ] ++ addlhdrs
         req = (ctx ^. baseRequest) { method = meth, requestHeaders = hdrs, requestBody = rbody }
         authreq = applyBasicAuth (ctx ^. basicusername) (ctx ^. basicpassword) req
-    resp <- lift (catchJust (matchStatusCodeException unauthorized401)
-                            (httpLbs req (ctx ^. httpManager))
-                            (\_ -> httpLbs authreq (ctx ^. httpManager)))
-    return resp
+    liftIO (catchJust (matchStatusCodeException unauthorized401)
+                      (httpLbs req (ctx ^. httpManager))
+                      (\_ -> httpLbs authreq (ctx ^. httpManager)))
 
 matchStatusCodeException :: Status -> HttpException -> Maybe ()
 matchStatusCodeException want (StatusCodeException s _ _)
@@ -94,87 +138,93 @@ matchStatusCodeException want (StatusCodeException s _ _)
     | otherwise = Nothing
 matchStatusCodeException _ _ = Nothing
 
-emptyBody :: RequestBody m
+emptyBody :: RequestBody
 emptyBody = RequestBodyLBS BL.empty
 
-xmlBody :: XML.Document -> RequestBody m
+xmlBody :: XML.Document -> RequestBody
 xmlBody = RequestBodyLBS . XML.renderLBS XML.def
 
-getOptions :: MonadResourceBase m => DAVState m ()
+getOptions :: MonadIO m => DAVT m ()
 getOptions = do
     optresp <- davRequest "OPTIONS" [] emptyBody
     let meths = (B.splitWith (==(fromIntegral . fromEnum) ',') . fromMaybe B.empty . lookup "Allow" . responseHeaders) optresp
     let cclass = (B.splitWith (==(fromIntegral . fromEnum) ',') . fromMaybe B.empty . lookup "DAV" . responseHeaders) optresp
-    modify (complianceClasses .~ cclass)
-    modify (allowedMethods .~ meths)
+    complianceClasses .= cclass
+    allowedMethods .= meths
 
-lockResource :: MonadResourceBase m => Bool -> DAVState m ()
+lockResource :: MonadIO m => Bool -> DAVT m ()
 lockResource nocreate = do
     let ahs' = [(hContentType, "application/xml; charset=\"utf-8\""), (mk "Depth", "0"), (mk "Timeout", "Second-300")]
     let ahs = if nocreate then (mk "If-Match", "*"):ahs' else ahs'
     lockresp <- davRequest "LOCK" ahs (xmlBody locky)
     let hdrtoken = (lookup "Lock-Token" . responseHeaders) lockresp
-    modify (lockToken .~ hdrtoken)
+    lockToken .= hdrtoken
 
-unlockResource :: MonadResourceBase m => DAVState m ()
+unlockResource :: MonadIO m => DAVT m ()
 unlockResource = do
     d <- get
     case _lockToken d of
         Nothing -> return ()
 	Just tok -> do let ahs = [(mk "Lock-Token", tok)]
                        _ <- davRequest "UNLOCK" ahs emptyBody
-                       modify (lockToken .~ Nothing)
+                       lockToken .= Nothing
 
-supportsLocking :: DAVContext a -> Bool
+supportsLocking :: DAVContext -> Bool
 supportsLocking = liftA2 (&&) ("LOCK" `elem`) ("UNLOCK" `elem`) . _allowedMethods
 
-supportsCalDAV :: DAVContext a -> Bool
+supportsCalDAV :: DAVContext -> Bool
 supportsCalDAV = ("calendar-access" `elem`) . _complianceClasses
 
-getPropsM :: MonadResourceBase m => DAVState m XML.Document
+getPropsM :: MonadIO m => DAVT m XML.Document
 getPropsM = do
     let ahs = [(hContentType, "application/xml; charset=\"utf-8\"")]
     propresp <- davRequest "PROPFIND" ahs (xmlBody propname)
-    return $ (XML.parseLBS_ def . responseBody) propresp
+    return $ (XML.parseLBS_ XML.def . responseBody) propresp
 
-getContentM :: MonadResourceBase m => DAVState m (Maybe B.ByteString, BL.ByteString)
+getContentM :: MonadIO m => DAVT m (Maybe B.ByteString, BL.ByteString)
 getContentM = do
     resp <- davRequest "GET" [] emptyBody
-    let ct = lookup (hContentType) (responseHeaders resp)
-    return $ (ct, responseBody resp)
+    let ct = lookup hContentType (responseHeaders resp)
+    return (ct, responseBody resp)
 
-putContentM :: MonadResourceBase m => (Maybe B.ByteString, BL.ByteString) -> DAVState m ()
+putContentM :: MonadIO m => (Maybe B.ByteString, BL.ByteString) -> DAVT m ()
 putContentM (ct, body) = do
     d <- get
-    let ahs' = fromMaybe [] (fmap (return . (,) (mk "If") . parenthesize) (d ^. lockToken))
-    let ahs = ahs' ++ fromMaybe [] (fmap (return . (,) (hContentType)) ct)
+    let ahs' = maybe [] (return . (,) (mk "If") . parenthesize) (d ^. lockToken)
+    let ahs = ahs' ++ maybe [] (return . (,) hContentType) ct
     _ <- davRequest "PUT" ahs (RequestBodyLBS body)
     return ()
 
-delContentM :: MonadResourceBase m => DAVState m ()
+delContentM :: MonadIO m => DAVT m ()
 delContentM = do
     _ <- davRequest "DELETE" [] emptyBody
     return ()
 
-moveContentM :: MonadResourceBase m => B.ByteString -> DAVState m ()
+moveContentM :: MonadIO m => B.ByteString -> DAVT m ()
 moveContentM newurl = do
     let ahs = [ (mk "Destination", newurl) ]
     _ <- davRequest "MOVE" ahs emptyBody
     return ()
 
-mkCol :: MonadResourceBase m => DAVState m ()
-mkCol = do
+mkCol' :: MonadIO m => DAVT m ()
+mkCol' = do
     _ <- davRequest "MKCOL" [] emptyBody
     return ()
 
+mkCol :: (MonadIO m, MonadBase IO m, MonadBaseControl IO m) => DAVT m Bool
+mkCol = catchJust
+        (matchStatusCodeException conflict409)
+        (mkCol' >> return True)
+        (\_ -> return False)
+
 parenthesize :: B.ByteString -> B.ByteString
 parenthesize x = B.concat ["(", x, ")"]
 
-putPropsM :: MonadResourceBase m => XML.Document -> DAVState m ()
+putPropsM :: MonadIO m => XML.Document -> DAVT m ()
 putPropsM props = do
     d <- get
     let ah' = (hContentType, "application/xml; charset=\"utf-8\"")
-    let ahs = ah':fromMaybe [] (fmap (return . (,) (mk "If") . parenthesize) (_lockToken d))
+    let ahs = ah':maybe [] (return . (,) (mk "If") . parenthesize) (_lockToken d)
     _ <- davRequest "PROPPATCH" ahs ((RequestBodyLBS . props2patch) props) -- FIXME: should diff and remove props from target
     return ()
 
@@ -200,66 +250,87 @@ props2patch = XML.renderLBS XML.def . patch . props . fromDocument
                    , "{DAV:}supportedlock"
                    ]
 
-caldavReportM :: MonadResourceBase m => DAVState m XML.Document
+caldavReportM :: MonadIO m => DAVT m XML.Document
 caldavReportM = do
     let ahs = [(hContentType, "application/xml; charset=\"utf-8\"")]
     calrresp <- davRequest "REPORT" ahs (xmlBody calendarquery)
-    return $ (XML.parseLBS_ def . responseBody) calrresp
+    return $ (XML.parseLBS_ XML.def . responseBody) calrresp
 
-getProps :: String -> B.ByteString -> B.ByteString -> Maybe Depth -> IO XML.Document
-getProps url username password md = withDS url username password md getPropsM
+getOptionsOnce :: MonadIO m => DAVT m ()
+getOptionsOnce = getOptions -- this should only happen once
 
-getPropsAndContent :: String -> B.ByteString -> B.ByteString -> IO (XML.Document, (Maybe B.ByteString, BL.ByteString))
-getPropsAndContent url username password = withDS url username password (Just Depth0) $ do
-    getOptions
+withLockIfPossible :: (MonadIO m, MonadBase IO m, MonadBaseControl IO m) => Bool -> DAVT m a -> DAVT m a
+withLockIfPossible nocreate f = do
+    getOptionsOnce
     o <- get
-    when (supportsLocking o) (lockResource True)
-    (do props <- getPropsM
-        body <- getContentM
-        return (props, body)) `finally` when (supportsLocking o) (unlockResource)
+    when (supportsLocking o) (lockResource nocreate)
+    f `finally` when (supportsLocking o) unlockResource
 
-putContent :: String -> B.ByteString -> B.ByteString -> (Maybe B.ByteString, BL.ByteString) -> IO ()
-putContent url username password b = withDS url username password Nothing $ do
-    getOptions
+withLockIfPossibleForDelete :: (MonadIO m, MonadBase IO m, MonadBaseControl IO m) => Bool -> DAVT m a -> DAVT m a
+withLockIfPossibleForDelete nocreate f = do
+    getOptionsOnce
     o <- get
-    when (supportsLocking o) (lockResource False)
-    putContentM b `finally` when (supportsLocking o) (unlockResource)
-
-putContentAndProps :: String -> B.ByteString -> B.ByteString -> (XML.Document, (Maybe B.ByteString, BL.ByteString)) -> IO ()
-putContentAndProps url username password (p, b) = withDS url username password Nothing $ do
-    getOptions
-    o <- get
-    when (supportsLocking o) (lockResource False)
-    (do putContentM b
-        putPropsM p) `finally` when (supportsLocking o) (unlockResource)
-
-deleteContent :: String -> B.ByteString -> B.ByteString -> IO ()
-deleteContent url username password = withDS url username password Nothing $ do
-    getOptions
-    o <- get
-    let lock = when (supportsLocking o) (lockResource False)
+    let lock = when (supportsLocking o) (lockResource nocreate)
     -- a successful delete destroys locks, so only unlock on error
-    let unlock = when (supportsLocking o) (unlockResource)
-    bracketOnError lock (\_ -> unlock) (\_ -> delContentM)
+    let unlock = when (supportsLocking o) unlockResource
+    bracketOnError lock (const unlock) (const f)
 
+{-# DEPRECATED getProps "This function will be removed in favor of getPropsM" #-}
+getProps :: String -> B.ByteString -> B.ByteString -> Maybe Depth -> IO XML.Document
+getProps url username password md = choke $ evalDAVT url $ do
+    setCreds username password
+    setDepth md
+    getPropsM
+
+{-# DEPRECATED getPropsAndContent "This function will be removed in favor of getPropsM and getContentM" #-}
+getPropsAndContent :: String -> B.ByteString -> B.ByteString -> IO (XML.Document, (Maybe B.ByteString, BL.ByteString))
+getPropsAndContent url username password = choke $ evalDAVT url $ do
+    setCreds username password
+    setDepth (Just Depth0)
+    withLockIfPossible True $ liftM2 (,) getPropsM getContentM
+
+{-# DEPRECATED putContent "This function will be removed in favor of putContentM" #-}
+putContent :: String -> B.ByteString -> B.ByteString -> (Maybe B.ByteString, BL.ByteString) -> IO ()
+putContent url username password b = choke $ evalDAVT url $ do
+    setCreds username password
+    withLockIfPossible False $ putContentM b
+
+{-# DEPRECATED putContentAndProps "This function will be removed in favor of putContentM and putPropsM" #-}
+putContentAndProps :: String -> B.ByteString -> B.ByteString -> (XML.Document, (Maybe B.ByteString, BL.ByteString)) -> IO ()
+putContentAndProps url username password (p, b) = choke $ evalDAVT url $ do
+    setCreds username password
+    withLockIfPossible False $ do putContentM b
+                                  putPropsM p
+
+{-# DEPRECATED deleteContent "This function will be removed in favor of delContentM" #-}
+deleteContent :: String -> B.ByteString -> B.ByteString -> IO ()
+deleteContent url username password = choke $ evalDAVT url $ do
+    setCreds username password
+    withLockIfPossibleForDelete False delContentM
+
+{-# DEPRECATED moveContent "This function will be removed in favor of moveContentM" #-}
 moveContent :: String -> B.ByteString -> B.ByteString -> B.ByteString -> IO ()
-moveContent url newurl username password = withDS url username password Nothing $
+moveContent url newurl username password = choke $ evalDAVT url $ do
+    setCreds username password
     moveContentM newurl
 
+{-# DEPRECATED caldavReport "This function will be removed in favor of caldavReportM" #-}
 caldavReport :: String -> B.ByteString -> B.ByteString -> IO XML.Document
-caldavReport url username password = withDS url username password (Just Depth1) $ caldavReportM
+caldavReport url username password = choke $ evalDAVT url $ do
+   setCreds username password
+   setDepth (Just Depth1)
+   caldavReportM
 
 -- | Creates a WebDAV collection, which is similar to a directory.
 --
 -- Returns False if the collection could not be made due to an intermediate
 -- collection not existing. (Ie, collection /a/b/c/d cannot be made until
 -- collection /a/b/c exists.)
+{-# DEPRECATED makeCollection "This function will be removed in favor of mkCol" #-}
 makeCollection :: String -> B.ByteString -> B.ByteString -> IO Bool
-makeCollection url username password = withDS url username password Nothing $
-    catchJust
-        (matchStatusCodeException conflict409)
-        (mkCol >> return True)
-        (\_ -> return False)
+makeCollection url username password = choke $ evalDAVT url $ do
+    setCreds username password
+    mkCol
 
 propname :: XML.Document
 propname = XML.Document (XML.Prologue [] Nothing []) root []
